/* Reference: "Stochastic Frontier Models with Errors in Variables: a GMM Approach" (in Chinese)   */
/*            by Yi-Yi Chen and Hung-Jen Wang, Taiwan Economic Review, 2009, 37 (1), 1-22.         */
/* contact: chenyiyi@mail.tku.edu.tw, wangh@ntu.edu.tw      */
/*                                             */
/* It estimates the following model:           */
/*                                             */
/*    y_i = b0 + X_i B + a z_i^* + v_i - u_i,  */
/*    z_i = z_i^* + e,                         */
/*    E(v) = 0, E(v^2) = sigma^2,              */
/*    u_i ~ N^+(mu, sigma_u^2),                */
/*                                             */
/*    where X_i is a vector of perfectly measured variable and z_i is the observed variable with error. */
/* Three GMM estimators are provided. GMM1 is just-identified, GMM2 uses one more moment condition, and GMM3 uses two more. */


new;
cls;
call sysstate(14,1e-128);
screen on;
_opusrbrk = 0;

 /* ----------------------------------------------------- */
 /* -------- Customization Begins ----------------------- */
 /* ----------------------------------------------------- */


noc = 1000;    @ number of obs @
nvar = 3;      @ number of variables in the dataset @
nx = 1;        @ number of perfectly measured regressors @
nestim = 3;    @ 1: GMM1; 2: GMM1 & GMM2; 3: GMM1 & GMM2 & GMM3  @

  /* -- Load Data and Define Variables -- */

load alldta[noc, nvar] = e:\\ajen\\ayiyi\\sf_gmm\\pub\\demodata.txt; @ the dataset should not contain var names @

y = alldta[1:noc,1];    @ the dep var @
x = alldta[1:noc,2];    @ the indep var, perfectly measured @
z = alldta[1:noc,3];    @ the indep var with measurement error @


  /* --- output file management ------- */

myout = "e:\\ajen\\ayiyi\\sf_gmm\\pub\\sfe_gmm.out";   @ save printout (coef, std.err., etc.) to this file @
jlmsout = "e:\\ajen\\ayiyi\\sf_gmm\\pub\\jlmsgmm.txt"; @ export the JLMS index to this file @
bcout = "e:\\ajen\\ayiyi\\sf_gmm\\pub\\bcgmm.txt";     @ export the BC index to this file @


  /* -- GMM initial values: Important!! -- */

GMM_4_true = 0.5; @ E(e^2); the variance of the measurement's error @
GMM_c_11 = 0.5;   @ the coeff of the var with  measurement error @
GMM_c_51 = 3;     @ sigma_u^2 @
GMM_c_61 = 1;     @ sigma_v^2 @
GMM_c_81 = 0.577;   @ mu_u / sigma_u @

  /* -- Error Trapping and Fixing ---- */

checkerr = 1;    @ 1: check, 0: don't check @
                 @ The results are more robust with the trapping enabled.@
                 @ It checks for errors in the Hessian matrix and tries to fix it.@
                 @ However, if the error cannot be fixed then the program halts. @
                 @ You decide whether you want to enable it! @

  /* ----- other parameters for GMM estimation; usually don'e need to change them ------ */

rndseed 135;
maxiter = 1000;  @-----set the maximum number of iterations-----@
maxsqez = 300;   @-----set the limit on the number of squeezes-----@
tol = 1e-6;      @-----set the tolerance on the minimization routine-----@


/* ------------------------------------------------ */
/* ---------- Customization Stop ------------------ */
/* ------------------------------------------------ */


  /* ---- other housekeeping ----- */

mygmmiter = zeros(1,nestim); @to take count of GMM iteration@

shell erase myout;
output file = ^myout reset;        @begin recording@

outwidth 132;

err_iscplxgwf = 0;
err_maxiter = 0;
err_cplxstd = 0;
err_mu2 = 0;
err_mu4 = 0;

n_obs = noc;


/* ------------------------------------- */
/* ---------  program begins ----------- */
/* ------------------------------------- */

    /* here the unobservable samples are used to
       calculate some initial value for gmm estimation
     */

    @ regress true_x on pm_x and constant @

    tmp_y = z; @true_x;@
    tmp_x = ones(n_obs,1)~x;  @ones(n_obs,1)~pm_x;@
    the_z1 = inv(tmp_x'tmp_x)* tmp_x'tmp_y;

    resid_z = tmp_y - tmp_x* the_z1;

    GMM_2_true = meanc(resid_z .* resid_z);
    GMM_3_true = meanc(resid_z .* resid_z .* resid_z);
    GMM_7_true = meanc(resid_z .* resid_z .* resid_z .* resid_z);


    @print "gmm_2: " GMM_2_true "  gmm_3: " GMM_3_true "  gmm_7:" GMM_7_true;@
    @print "gmm_4: " GMM_4_true;@


/* ------------------------------------------------------ */


    @-----The main part of the program-----@

    icept = ones(noc,1);
        xa = x;          /* x=pm_x, the indep var without measurement error */
        ix = icept~xa;   /* include the intercept */

    @-----demean-----@
    muy = (invpd(moment(ix,0)))*(ix'y);
    muz = (invpd(moment(ix,0)))*(ix'z);

    y_d = y - ix*muy;
    z_d = z - ix*muz;


    @-----construct the moments for subsequent estimation-----@
    y2_d = y_d^2;
    z2_d = z_d^2;
    z3_d = z_d^3;
    z4_d = z_d^4;
    yz_d = y_d.*z_d;
    y2z_d = y2_d.*z_d;
    yz2_d = y_d.*z2_d;
    y3_d = y_d^3;
    y2z2_d = y2_d.*z2_d;
    y3z_d = y3_d.*z_d;
    yz3_d = y_d.*z3_d;
    y4_d = y_d^4;

    mom = { y2_d  yz_d  z2_d  y2z_d  yz2_d  y3_d yz3_d y4_d y3z_d y2z2_d };


    Ey2_d = meanc(y2_d);
    Eyz_d = meanc(yz_d);
    Ez2_d = meanc(z2_d);
    Ez3_d = meanc(z3_d);
    Ez4_d = meanc(z4_d);
    Ey2z_d = meanc(y2z_d);
    Eyz2_d = meanc(yz2_d);
    Ey3_d = meanc(y3_d);
    Ey2z2_d = meanc(y2z2_d);
    Ey3z_d = meanc(y3z_d);
    Eyz3_d = meanc(yz3_d);
    Ey4_d = meanc(y4_d);

    Emom = { Ey2_d  Eyz_d  Ez2_d  Ey2z_d  Eyz2_d  Ey3_d Eyz3_d Ey4_d Ey3z_d Ey2z2_d };



    Exz = ix'z/noc; @ make the moments for the second stage of the
            partialling estimator@

    @The true statistics of regressors@
    Ey = meanc(y);
    Ez = meanc(z);

    Ey2 = meanc(y^2);
    Ez2 = meanc(z^2);

    y2z = (y^2) .*z;
    yz2 = y .*(z^2);
    Ey2z = meanc(y2z);
    Eyz2 = meanc(yz2);

    if nx > 0;
        Ex = meanc(x);
        Ex2 = meanc(x^2);
    endif;

    varz = ((z-Ez)'*(z-Ez))/noc;

    /*make the moments for the correct standard errors for beta*/

    Ey2_x = meanc(y2_d.*ix);
    Eyz_x = meanc(yz_d.*ix);
    Ez2_x = meanc(z2_d.*ix);
    Ey2z_x = meanc(y2z_d.*ix);
    Eyz2_x = meanc(yz2_d.*ix);
    Ey3_x = meanc(y3_d.*ix);
    Ez3_x = meanc(z3_d.*ix);

    inExx = invpd(moment(ix,0)/noc);

    xy_d = ix.*y_d;
    xz_d = ix.*z_d;


/* ------ The GMM Estimation ----- */


        {exit_status, asave, dsave, csave, usave, sigmau2save, sigmav2save,
            corr_jlms, corr_bc, scorr_jlms, scorr_bc, gmm_jlms, gmm_bc, obj1, obj2, oidsave} = gmm_est(noc, nx);

        @print "   GMM Estimation Finished.";@

        screen off;
          export(gmm_jlms, jlmsout, 0); @ export the JLMS index @
          export(gmm_bc, bcout, 0);     @ export the BC index @
        screen on;



            @GMM results@
            print;
            print " ***********************************************************";
            print " ************* GMM Estimation Results: *****************";
            print " ***********************************************************";
            print " The number of iterations (GMM): " mygmmiter;
            print " intercept (1st line) and standard error (2nd line) " asave;
            print " the perfectly measured coefficient (1st line) and standard error (2nd line; alternatively) = " dsave;
            print " the imperfectly measured coefficient (1st line) and standard error (2nd line) = " csave;
            print " the estimated mean of u dist (i.e., mu) = " usave;
            print " the variance of sigmau^2 = " sigmau2save;
            print " the variance of sigmav^2 = " sigmav2save;
            @print "corr_jlms|scorr_jlms is " corr_jlms|scorr_jlms;@
            @print "corr_bc|scorr_bc is " corr_bc|scorr_bc;@
            print "Over ID test:" oidsave;
            print "The JLMS and BC index are saved in separated output files.";
            print;



  /*
            gmm1 = asave[1,1]~dsave[1,1]~csave[1,1]~usave[1,1]~sigmau2save[1,1]~sigmav2save[1,1];
            gmm2 = asave[1,2]~dsave[1,2]~csave[1,2]~usave[1,2]~sigmau2save[1,2]~sigmav2save[1,2];
            gmm3 = asave[1,3]~dsave[1,3]~csave[1,3]~usave[1,3]~sigmau2save[1,3]~sigmav2save[1,3];
  */




/* --------------------------------------------- */



/*
  The GMM estimators
  The gmm_est (and other function used by it) takes all global variables
  defined in Emom, conduct the estimation, and passback the
  estimation results.

 */

proc (16) = gmm_est(noc, nx);

    local estim, neq, nc, c, ff, w, f, iter, dc, g, obj, gwg, oidv;
    local gwf, c_new, mm, vc, sqz, stderr, inflnc, gee, bigphi, avar, dstd;
    local omega, phidel, astdd, phidela, ustd, allsave;
    local asave, csave, dsave, sigmav2save, sigmau2save, usave, iasave, icsave, idsave, iusave;
    local qq, muu, comres, mustar, comsigma2star, comsigmastar, jlmsGMM, bcGMM;
    local exit_status, tmp1, tmp2, tmp3;
    local corr_jlms, corr_bc, scorr_jlms, scorr_bc, obj1, obj2, t, df, pvt, nww, test, nov;
    local l, wl, dl, fl, gl, objl, gwgl, gwfl, onemore, l_new, vl;
    local gmm1_result;
    local corig, stdorg;

    exit_status = 0;

    obj1=zeros(3,1);
    obj2=zeros(3,1);
    @-----parameters-----@
    asave = zeros(2, nestim); @intercept@
    csave = zeros(2, nestim);  @ coefficient on imperfectly measured regressors@
    oidsave = zeros(2, nestim); @ overidentification test@

    if nx > 0;
        dsave = zeros(2*nx, nestim); @coefficients on perfectly measured regressors@

        @allsave = zeros(1,2*nestim*(4+nx));@
    endif;
    sigmav2save = zeros(2, nestim);
    sigmau2save = zeros(2, nestim);
    muu = zeros(1, nestim);
    usave = zeros(2, nestim);

    comres = zeros(noc, nestim);
    mustar = zeros(noc, nestim);
    comsigma2star = zeros(1, nestim);
    comsigmastar = zeros(1, nestim);
    jlmsGMM = zeros(noc, nestim);
    bcGMM = zeros(noc, nestim);
    corr_jlms = zeros(1, nestim);
    scorr_jlms = zeros(1, nestim);
    corr_bc = zeros(1, nestim);
    scorr_bc = zeros(1, nestim);

    @-----Influence Functions-----@
    iasave = zeros(noc, nestim);
    icsave = zeros(noc, nestim);
    if nx > 0;
        idsave = zeros(noc*nx, nestim);
    endif;
    iusave = zeros(noc, nestim);

    estim = 1;
    do while estim <= nestim;

        if estim==1;  /* HJW need to change this */
            nc = 8;
            neq = 8;  /* the number of moments for GMM1 */
        elseif estim==2;
            nc = 8;
            neq = 9;  /* the number of moments for GMM2 */
        elseif estim ==3;
            nc = 8;
            neq = 10; /* the number of moments for GMM3 */
        endif;


        @-----input the starting values-----@

        c = zeros(nc,1);  /* HJW need to take care of nc */


        /*
        @use observable sample moments@
        c[2,1] = ln(Eyz_d/c[1,1]); @E(z^*-\bar{z})^2@
        c[3,1] = Eyz2_d/c[1,1]; @E(z^*-\bar{z})^3@
        c[4,1] = ln(Ez2_d - exp(c[2,1])); @E(e^2)@
        c[7,1] = ln((Eyz3_d - (3*c[1,1]*exp(c[2,1])*exp(c[4,1])))/c[1,1]); @E(z^*-\bar{z})^4@
        */

        if estim == 1;
            @use unobservable sample moments@  /* HJW: the c vector contains parameters of the model. Leave it as it is now. */
            c[2,1] = ln(GMM_2_true); @E(z^*-\bar{z})^2@
            c[3,1] = GMM_3_true; @E(z^*-\bar{z})^3@
            c[4,1] = ln(GMM_4_true); @E(e^2)@
            c[7,1] = ln(GMM_7_true); @E(z^*-\bar{z})^4@


            c[1,1] = GMM_c_11; @(Ey2z_d)/(Eyz2_d);@ @coef of measurement error variable@
            c[5,1] = ln(GMM_c_51);  @one parameter from f(u), \sigma_u^2@
            c[6,1] = ln(GMM_c_61); @ E(v^2)@
            c[8,1] = GMM_c_81; @the other parameter from f(u), \mu_u/\sigma_u@
        else;
            c=gmm1_result;
        endif;

        @----------Apply Nonliear Equation---------------@

        @-----the program creates the weighting matrix-----@
        if estim >= 1;
            ff = optw(c, mom, Emom, estim, neq);
            w = invit(moment(ff,0)./noc);
            f = deff(c, Emom, estim, neq);

            @-----start of iteration loop-----@

            iter = 1;
            dc = 1;
            do until abs(dc) < tol;

                if estim == 1;
                    g=grad1(c, neq);
                elseif estim == 2;
                    g=grad2(c, neq);
                elseif estim == 3;
                    g=grad3(c, neq);
                endif;

                obj = f*w*f';

                gwg= g'w*g;               @ This uses the GAUSS-NEWTON method
                                         to compute full step dc. @
                gwf = g'w*f';

                if(iscplx(gwf) or iscplx(gwg));
                    exit_status = 1;
                    err_iscplxgwf = err_iscplxgwf + 1;
                    goto direct_return;
                endif;

                if checkerr == 1;
                  dc = solvit(gwf,gwg, estim);  /* HJW: need further investigation */
                endif;

                if maxsqez > 0;              @ This jumps to the subroutine that
                                                adjusts the step length. @
                    { c_new,sqz } = squeeze(c,dc, estim, neq, w, obj);
                else;
                    c_new=c - dc;
                endif;


                dc=c_new-c;                     @ Update variables for the next iteration. @
                c=c_new;

                /*--------------Print out the results for the current iteration------------*/

              /* HJW: not necessary
                format 14,6;
                mm=1;
                do until mm > nc;
                    $"  ";; c[mm,1];; "  ";; dc[mm,1];
                    mm=mm+1;
                endo;
              */

                iter=iter +1;
                if iter >= maxiter;                     @ Quit iterating if necessary. @

                    @goto escape;@ @comment out by HJW@
                    @following two lines added by HJW@
                    exit_status = 1;
                    err_maxiter = err_maxiter + 1;
                    goto direct_return;
                endif;
            endo;

            @escape:@
        endif;

  @ HJW: Here, the coefficient vector c has beem computed. @

        corig = c;
        corig[2,1] = exp(c[2,1]);
        corig[4,1] = exp(c[4,1]);
        corig[5,1] = exp(c[5,1]);
        corig[6,1] = exp(c[6,1]);
        corig[7,1] = exp(c[7,1]);


        if estim == 1;
            gmm1_result = c;
        endif;

        mygmmiter[1, estim] = iter;

        ff = optw(c,mom,Emom, estim, neq);
        f = deff(c,Emom, estim, neq);

        w = invit(moment(ff,0)./noc);

        if estim == 1;
            g=grad1(c, neq);
        elseif estim == 2;
            g=grad2(c, neq);
        elseif estim == 3;
            g=grad3(c, neq);
        endif;

        oidv = noc*f*w*f';
        df = neq - nc;

        test = cdfchic(oidv,df); @the p-value of the chi-square distribution@

        oidsave[1, estim] = oidv;
        oidsave[2, estim] = test;

        gwg = g'w*g;
        vc = invit(gwg)/noc;
        stderr = sqrt(diag(vc));

        inflnc = -noc*vc*g'w*ff';

        if(iscplx(stderr));
            exit_status = 1;
            err_cplxstd = err_cplxstd + 1;
            goto direct_return;
        endif;

        @the second central moment of u should not be negative@
        if(mu2(c[5,1],c[8,1])<= 0);
            exit_status = 1;
            err_mu2 = err_mu2 + 1;
            goto direct_return;
        endif;

        @the 4th central moment of u should not be negative@
        if(mu4(c[5,1],c[8,1])<= 0);
            exit_status = 1;
            err_mu4 = err_mu4 + 1;
            goto direct_return;
        endif;

        csave[1, estim] = c[1,1];
        csave[2, estim] = stderr[1,1];
        icsave[1:noc, estim] = inflnc[1,.]';

        sigmav2save[1, estim] = exp(c[6,1]);
/* HJW
        print "estim is " estim;
        print "stderr is " stderr;
*/
        sigmav2save[2, estim] = exp(c[6,1])*stderr[6,1];
        sigmau2save[1, estim] = exp(c[5,1]);
        sigmau2save[2, estim] = exp(c[5,1])*stderr[5,1];
        muu[1, estim] = c[8,1];

        stdorg = stderr;
        stdorg[2,1] = exp(c[2,1])*stderr[2,1];
        stdorg[4,1] = exp(c[4,1])*stderr[4,1];
        stdorg[5,1] = exp(c[5,1])*stderr[5,1];
        stdorg[6,1] = exp(c[6,1])*stderr[6,1];
        stdorg[7,1] = exp(c[7,1])*stderr[7,1];

        /* make the estimated coefficient of perfectly measured regressors */

        if nx > 0;
            for qq(1, nx, 1);
                dsave[(qq*2-1), estim] = muy[qq+1,1] - c[1,1]*muz[qq+1,1];
            endfor;

            gee = zeros(((nx+1)*2+1),nx);

            gee[2:(nx+1),1:nx] = eye(nx);
            gee[(nx+3):((nx+1)*2),1:nx] = -c[1,1]*eye(nx);

            for qq(1, nx, 1);
                gee[((nx+1)*2+1),qq] = -muz[qq+1,1];
            endfor;

            @---Correct standard error and influence function---@

            bigphi = ( (inExx*xy_d')|(inExx*xz_d') )|(-inflnc[1,.]);
            avar = moment(bigphi',0)./noc^2;
            dstd = sqrt(diag(gee'avar*gee));
            omega = gee'avar*gee;
            phidel = -bigphi'gee;

            for qq(1, nx, 1);
                dsave[qq*2,estim] = dstd[qq,1];
            endfor;

            for qq(1, nx, 1);
                idsave[((qq-1)*noc+1):(qq*noc),estim] = phidel[.,qq];
            endfor;

        endif;

        /* make the stupid intercept */
        asave[1, estim] = muy[1,1] - c[1,1]*muz[1,1] + sqrt(exp(c[5,1]))*(c[8,1]+lamda(c[8,1]));

        gee = zeros(((nx+1)*2+3),1);

        gee[1,1] = 1;
        gee[(nx+2),1] = -c[1,1];
        gee[((nx+1)*2+1),1] = -muz[1,1];
        gee[((nx+1)*2+2),1] = (0.5*(1/sqrt(exp(c[5,1]))))*(c[8,1]+lamda(c[8,1]));
        gee[((nx+1)*2+3),1] = sqrt(exp(c[5,1]))*(1-(lamda(c[8,1])*(c[8,1]+lamda(c[8,1]))));
        @---Correct standard error and influence function---@

        bigphi = ( (inExx*xy_d')|(inExx*xz_d') )|(-inflnc[1,.])|(-inflnc[5,.])|(-inflnc[8,.]);
        avar = moment(bigphi',0)./noc^2;
        astdd = sqrt(diag(gee'avar*gee));
        phidela = -bigphi'gee;

        asave[2,estim] = astdd[1,1];
        iasave[1:noc,estim] = phidela[.,1];

        /* make the mean of inefficiency */
        usave[1, estim] =  sqrt(exp(c[5,1]))*c[8,1];

        gee = zeros(2,1);

        gee[1,1] = -0.5*(1/sqrt(exp(c[5,1])))*c[8,1];
        gee[2,1] = -sqrt(exp(c[5,1]));
        @---Correct standard error and influence function---@

        bigphi = (-inflnc[5,.])|(-inflnc[8,.]);
        avar = moment(bigphi',0)./noc^2;
        ustd = sqrt(diag(gee'avar*gee));
        phidela = -bigphi'gee;

        usave[2,estim] = ustd[1,1];
        iusave[1:noc,estim] = phidela[.,1];

          /*
          @already check, marked it up. HML@
        if(sigmav2save[1, estim] <= 0 or sigmau2save[1, estim] <= 0);
            exit_status = 1;
            goto direct_return;
        endif;
        */


        comres[., estim] = y - icept*asave[1,estim] - x*dsave[1:nx, estim] - z*csave[1, estim];
        mustar[., estim] = (-comres[., estim]*sigmau2save[1, estim] + sigmav2save[1, estim]*muu[1, estim])/(sigmav2save[1, estim]+sigmau2save[1, estim]);
        comsigma2star[1, estim] = (sigmav2save[1, estim]*sigmau2save[1, estim])/(sigmav2save[1, estim]+sigmau2save[1, estim]);
        comsigmastar[1, estim] = sqrt(comsigma2star[1, estim]);

/* HJW
        print "sigmav2save is " sigmav2save;
        print "sigmau2save is " sigmau2save;
        print "comsigma2star is " comsigma2star;
*/

        @debug1 =  mustar[., estim]/comsigmastar[1, estim];@
        tmp1 = lamda(mustar[., estim]/comsigmastar[1, estim]);
        jlmsGMM[., estim] = comsigmastar[1, estim]* tmp1 + mustar[1:noc, estim];
        tmp1 = exp( - mustar[., estim] + icept*comsigma2star[1, estim]/2);
        tmp2 = cdfn(mustar[., estim]/comsigmastar[1, estim] - comsigmastar[1, estim]);
        tmp3 = cdfn(mustar[., estim]/comsigmastar[1, estim]);
        bcGMM[., estim] =  tmp1 .* tmp2 ./ tmp3;

/*
        tmp1 =  corrx(true_jlms~jlmsGMM[., estim]);
        corr_jlms[estim] = tmp1[1,2];
        {tmp1, tmp2} = spearmans_corrx(true_jlms, jlmsGMM[., estim]);
        scorr_jlms[estim] = tmp1;
*/
        scorr_jlms[estim] = 1; /* just to fool the program */

/*
        tmp1 = corrx(true_bc~bcGMM[., estim]);
        corr_bc[estim]=tmp1[1,2];
        {tmp1, tmp2} = spearmans_corrx(true_bc, bcGMM[., estim]);
        scorr_bc[estim] = tmp1;
*/
        scorr_bc[estim] = 1; /* just to fool the program */

        obj1[estim] = obj;

        estim = estim + 1;
    endo;

    direct_return:


    obj2=obj1*noc;

    retp(exit_status, asave, dsave, csave, usave, sigmau2save, sigmav2save,
         corr_jlms, corr_bc, scorr_jlms, scorr_bc, jlmsGMM, bcGMM, obj1, obj2, oidsave);
endp;




proc lamda(b);
    local h;
    h = pdfn(b)./cdfn(b);
    retp(h);
endp;

proc mu2(a,b);
    local h;
    h = (exp(a))*(-(lamda(b)^2)-(b*lamda(b))+1);
    retp(h);
endp;

proc mu2_a(a,b);
    local h;
    h = (exp(a))*(-(lamda(b)^2)-(b*lamda(b))+1);
    retp(h);
endp;

proc mu2_b(a,b);
    local h;
    h = (exp(a))*((-(2*lamda(b))-b)*(-lamda(b)*(b+lamda(b)))-lamda(b));
    retp(h);
endp;

proc mu3(a,b);
    local h;
    h = sqrt(exp(a)^3)*lamda(b)*((2*(lamda(b)^2))+(3*b*lamda(b))+(b^2)-1);
    retp(h);
endp;

proc mu3_a(a,b);
    local h;
    h = (1.5)*sqrt(exp(a)^3)*lamda(b)*((2*(lamda(b)^2))+(3*b*lamda(b))+(b^2)-1);
    retp(h);
endp;

proc mu3_b(a,b);
    local h;
    h =
        (sqrt(exp(a)^3))*((6*(lamda(b)^2)+(6*b*lamda(b))+(b^2)-1)*(-lamda(b)*(b+lamda(b)))+
        (3*(lamda(b)^2)+(2*b*lamda(b))));
    retp(h);
endp;

proc mu4(a,b);
    local h;
    h =
        -(exp(a)^2)*((3*(lamda(b)^4))+(6*b*(lamda(b)^3))+(4*(b^2)*(lamda(b)^2))+
        ((b^3)*lamda(b))+(3*b*lamda(b))+(2*(lamda(b)^2))-3);
    retp(h);
endp;

proc mu4_a(a,b);
    local h;
    h = -2*(exp(a)^2)*((3*(lamda(b)^4))+(6*b*(lamda(b)^3))+(4*(b^2)*(lamda(b)^2))+
        ((b^3)*lamda(b))+(3*b*lamda(b))+(2*(lamda(b)^2))-3);
    retp(h);
endp;

proc mu4_b(a,b);
    local h;
    h = -(exp(a)^2)*((12*(lamda(b)^3)+(18*b*(lamda(b)^2))+
        (8*(b^2)*lamda(b))+(b^3)+(3*b)+
        (4*(lamda(b)))*(-lamda(b)*(b+lamda(b))))+
        (6*(lamda(b)^3)+
        (8*(b)*(lamda(b)^2))+(3*(b^2)*lamda(b))+(3*lamda(b))));
    retp(h);
endp;


proc (2)=spearmans_corrx(x,y);
    local n, z, xrank, yrank, rs, t;
    n = rows(x);
    z = sortc(x~y,1);
    xrank = sortind(z[.,1]);
    yrank = sortind(z[.,2]);
    rs = 1-6*(xrank-yrank)'(xrank-yrank)/(n*(n^2-1));
    t = rs*sqrt((n-2)/(1-rs^2));
    retp(rs,t);
endp;


@-----subroutines to define the f factor for the optimal weighting
matrix-----@

proc optw(a, mmm, ooo, estim, neq);
    local f, ei;

    f = zeros(noc, neq);

    f[.,1] = varget(mmm[1]) - varget(ooo[1]);
    f[.,2] = varget(mmm[2]) - varget(ooo[2]);
    f[.,3] = varget(mmm[3]) - varget(ooo[3]);
    f[.,4] = varget(mmm[4]) - varget(ooo[4]);
    f[.,5] = varget(mmm[5]) - varget(ooo[5]);
    f[.,6] = varget(mmm[6]) - varget(ooo[6]);
    f[.,7] = varget(mmm[7]) - varget(ooo[7]);
    f[.,8] = varget(mmm[8]) - varget(ooo[8]);
    if estim > 1;
        f[.,9] = varget(mmm[9]) - varget(ooo[9]);
    endif;
    if estim > 2;
        f[.,10] = varget(mmm[10]) - varget(ooo[10]);
    endif;

    @-----makes the standard error adjustment-----@
    ei = zeros(noc, neq);
    ei[.,4] = (-2*Eyz_x'inExx*xy_d' - Ey2_x'inExx*xz_d')';
    ei[.,5] = (-Ez2_x'inExx*xy_d' - 2*Eyz_x'inExx*xz_d')';
    ei[.,6] = (-3*Ey2_x'inExx*xy_d')';
    ei[.,7] = (-Ez3_x'inExx*xy_d' - 3*Eyz2_x'inExx*xz_d')';
    ei[.,8] = (-4*Ey3_x'inExx*xy_d')';
    if estim > 1;
        ei[.,9] = (-3*Ey2z_x'inExx*xy_d' - Ey3_x'inExx*xz_d')';
    endif;

    if estim > 2;
        ei[.,10] = (-2*Eyz2_x'inExx*xy_d' -2*Ey2z_x'inExx*xz_d')';
    endif;
    f = f + ei;

    retp(f);
    clear f;
endp;
@----------------------------------------------------------------@
@--------------------------------------------------------------@
proc deff(a, mmm, estim, neq);
    local f;

    f = zeros(1,neq);

    f[.,1] = varget(mmm[1]) - (a[1,1]^2)*exp(a[2,1]) - exp(a[6,1]) - mu2(a[5,1],a[8,1]);
    f[.,2] = varget(mmm[2]) - a[1,1]*exp(a[2,1]);
    f[.,3] = varget(mmm[3]) - exp(a[2,1]) - exp(a[4,1]);
    f[.,4] = varget(mmm[4]) - (a[1,1]^2)*a[3,1];
    f[.,5] = varget(mmm[5]) - a[1,1]*a[3,1];
    f[.,6] = varget(mmm[6]) - (a[1,1]^3)*a[3,1] + mu3(a[5,1],a[8,1]);
    f[.,7] = varget(mmm[7]) - (a[1,1]*exp(a[7,1])) - (3*a[1,1]*exp(a[2,1])*exp(a[4,1]));
    f[.,8] = varget(mmm[8]) - (a[1,1]^4)*exp(a[7,1]) - 6*(a[1,1]^2)*exp(a[2,1])*
        (exp(a[6,1]) + mu2(a[5,1],a[8,1])) - 3*((exp(a[6,1]))^2) -
        (6*exp(a[6,1])* mu2(a[5,1],a[8,1])) - mu4(a[5,1],a[8,1]);
    if estim > 1;
        f[.,9] = varget(mmm[9]) - (a[1,1]^3)*exp(a[7,1]) -
            3*a[1,1]*exp(a[2,1])*(exp(a[6,1]) +  mu2(a[5,1],a[8,1]));
    endif;
    if estim > 2;
        f[.,10] = varget(mmm[10]) -
            (a[1,1]^2)*(exp(a[7,1])+(exp(a[2,1])*exp(a[4,1]))) - (exp(a[2,1]) + exp(a[4,1]))*
            (exp(a[6,1]) + mu2(a[5,1],a[8,1]));
    endif;

    retp(f);
endp;


@---------------------------------------------------------------@

@-----subroutine to compute gradient matrix-----@


proc grad1(a, neq);

    local dvdc1, dvdc2, dvdc3, dvdc4, dvdc5, dvdc6, dvdc7, dvdc8;

    dvdc1 = zeros(neq,1);
    dvdc2 = zeros(neq,1);
    dvdc3 = zeros(neq,1);
    dvdc4 = zeros(neq,1);
    dvdc5 = zeros(neq,1);
    dvdc6 = zeros(neq,1);
    dvdc7 = zeros(neq,1);
    dvdc8 = zeros(neq,1);

    dvdc1[1,1]= -2*a[1,1]*exp(a[2,1]);
    dvdc1[2,1]= -exp(a[2,1]);
    dvdc1[3,1]= 0;
    dvdc1[4,1]= -2*a[1,1]*a[3,1];
    dvdc1[5,1]= -a[3,1];
    dvdc1[6,1]= -3*(a[1,1]^2)*a[3,1];
    dvdc1[7,1]= -exp(a[7,1])-3*exp(a[2,1])*exp(a[4,1]);
    dvdc1[8,1]= -(4*(a[1,1]^3)*exp(a[7,1]))-12*a[1,1]*exp(a[2,1])*(exp(a[6,1])+
    mu2(a[5,1],a[8,1]));

    dvdc2[1,1]= -(a[1,1]^2)*exp(a[2,1]);
    dvdc2[2,1]= -a[1,1]*exp(a[2,1]);
    dvdc2[3,1]= -exp(a[2,1]);
    dvdc2[4,1]= 0;
    dvdc2[5,1]= 0;
    dvdc2[6,1]= 0;
    dvdc2[7,1]= -3*a[1,1]*exp(a[4,1])*exp(a[2,1]);
    dvdc2[8,1]= -(6*(a[1,1]^2)*(exp(a[6,1]) + mu2(a[5,1],a[8,1])))*exp(a[2,1]);

    dvdc3[1,1]= 0;
    dvdc3[2,1]= 0;
    dvdc3[3,1]= 0;
    dvdc3[4,1]= -(a[1,1]^2);
    dvdc3[5,1]= -a[1,1];
    dvdc3[6,1]= -(a[1,1]^3);
    dvdc3[7,1]= 0;
    dvdc3[8,1]= 0;

    dvdc4[1,1]= 0;
    dvdc4[2,1]= 0;
    dvdc4[3,1]= -exp(a[4,1]);
    dvdc4[4,1]= 0;
    dvdc4[5,1]= 0;
    dvdc4[6,1]= 0;
    dvdc4[7,1]= -3*a[1,1]*exp(a[2,1])*exp(a[4,1]);
    dvdc4[8,1]= 0;

    dvdc5[1,1]= -mu2_a(a[5,1],a[8,1]);
    dvdc5[2,1]= 0;
    dvdc5[3,1]= 0;
    dvdc5[4,1]= 0;
    dvdc5[5,1]= 0;
    dvdc5[6,1]= mu3_a(a[5,1],a[8,1]);
    dvdc5[7,1]= 0;
    dvdc5[8,1]= -6*(((a[1,1]^2)*exp(a[2,1]))+exp(a[6,1]))*mu2_a(a[5,1],a[8,1])
        -mu4_a(a[5,1],a[8,1]);

    dvdc6[1,1]= -exp(a[6,1]);
    dvdc6[2,1]= 0;
    dvdc6[3,1]= 0;
    dvdc6[4,1]= 0;
    dvdc6[5,1]= 0;
    dvdc6[6,1]= 0;
    dvdc6[7,1]= 0;
    dvdc6[8,1]= -6*(mu2(a[5,1],a[8,1]))*exp(a[6,1])-(6*(a[1,1]^2)*exp(a[2,1])*exp(a[6,1]))-6*exp(2*a[6,1]);


    dvdc7[1,1]= 0;
    dvdc7[2,1]= 0;
    dvdc7[3,1]= 0;
    dvdc7[4,1]= 0;
    dvdc7[5,1]= 0;
    dvdc7[6,1]= 0;
    dvdc7[7,1]= -a[1,1]*exp(a[7,1]);
    dvdc7[8,1]= -(a[1,1]^4)*exp(a[7,1]);

    dvdc8[1,1]= -mu2_b(a[5,1],a[8,1]);
    dvdc8[2,1]= 0;
    dvdc8[3,1]= 0;
    dvdc8[4,1]= 0;
    dvdc8[5,1]= 0;
    dvdc8[6,1]= mu3_b(a[5,1],a[8,1]);
    dvdc8[7,1]= 0;
    dvdc8[8,1]= -6*(((a[1,1]^2)*exp(a[2,1]))+exp(a[6,1]))*mu2_b(a[5,1],a[8,1])-
        mu4_b(a[5,1],a[8,1]);

    retp(dvdc1~dvdc2~dvdc3~dvdc4~dvdc5~dvdc6~dvdc7~dvdc8);
endp;

proc grad2(a, neq);

    local dvdc1, dvdc2, dvdc3, dvdc4, dvdc5, dvdc6, dvdc7, dvdc8;

    dvdc1 = zeros(neq,1);
    dvdc2 = zeros(neq,1);
    dvdc3 = zeros(neq,1);
    dvdc4 = zeros(neq,1);
    dvdc5 = zeros(neq,1);
    dvdc6 = zeros(neq,1);
    dvdc7 = zeros(neq,1);
    dvdc8 = zeros(neq,1);

    dvdc1[1,1]= -2*a[1,1]*exp(a[2,1]);
    dvdc1[2,1]= -exp(a[2,1]);
    dvdc1[3,1]= 0;
    dvdc1[4,1]= -2*a[1,1]*a[3,1];
    dvdc1[5,1]= -a[3,1];
    dvdc1[6,1]= -3*(a[1,1]^2)*a[3,1];
    dvdc1[7,1]= -exp(a[7,1])-3*exp(a[2,1])*exp(a[4,1]);
    dvdc1[8,1]= -(4*(a[1,1]^3)*exp(a[7,1]))-12*a[1,1]*exp(a[2,1])*(exp(a[6,1])+
    mu2(a[5,1],a[8,1]));
    dvdc1[9,1]= -3*(a[1,1]^2)*exp(a[7,1])-3*exp(a[2,1])*(exp(a[6,1])+mu2(a[5,1],a[8,1]));

    dvdc2[1,1]= -(a[1,1]^2)*exp(a[2,1]);
    dvdc2[2,1]= -a[1,1]*exp(a[2,1]);
    dvdc2[3,1]= -exp(a[2,1]);
    dvdc2[4,1]= 0;
    dvdc2[5,1]= 0;
    dvdc2[6,1]= 0;
    dvdc2[7,1]= -3*a[1,1]*exp(a[4,1])*exp(a[2,1]);
    dvdc2[8,1]= -(6*(a[1,1]^2)*(exp(a[6,1]) + mu2(a[5,1],a[8,1])))*exp(a[2,1]);
    dvdc2[9,1]= -(3*a[1,1]*(exp(a[6,1])+mu2(a[5,1],a[8,1])))*exp(a[2,1]);

    dvdc3[1,1]= 0;
    dvdc3[2,1]= 0;
    dvdc3[3,1]= 0;
    dvdc3[4,1]= -(a[1,1]^2);
    dvdc3[5,1]= -a[1,1];
    dvdc3[6,1]= -(a[1,1]^3);
    dvdc3[7,1]= 0;
    dvdc3[8,1]= 0;
    dvdc3[9,1]= 0;

    dvdc4[1,1]= 0;
    dvdc4[2,1]= 0;
    dvdc4[3,1]= -exp(a[4,1]);
    dvdc4[4,1]= 0;
    dvdc4[5,1]= 0;
    dvdc4[6,1]= 0;
    dvdc4[7,1]= -3*a[1,1]*exp(a[2,1])*exp(a[4,1]);
    dvdc4[8,1]= 0;
    dvdc4[9,1]= 0;

    dvdc5[1,1]= -mu2_a(a[5,1],a[8,1]);
    dvdc5[2,1]= 0;
    dvdc5[3,1]= 0;
    dvdc5[4,1]= 0;
    dvdc5[5,1]= 0;
    dvdc5[6,1]= mu3_a(a[5,1],a[8,1]);
    dvdc5[7,1]= 0;
    dvdc5[8,1]= -6*(((a[1,1]^2)*exp(a[2,1]))+exp(a[6,1]))*mu2_a(a[5,1],a[8,1])
        -mu4_a(a[5,1],a[8,1]);
    dvdc5[9,1]= -3*a[1,1]*exp(a[2,1])*mu2_a(a[5,1],a[8,1]);

    dvdc6[1,1]= -exp(a[6,1]);
    dvdc6[2,1]= 0;
    dvdc6[3,1]= 0;
    dvdc6[4,1]= 0;
    dvdc6[5,1]= 0;
    dvdc6[6,1]= 0;
    dvdc6[7,1]= 0;
    dvdc6[8,1]= -6*mu2(a[5,1],a[8,1])*exp(a[6,1])-(6*(a[1,1]^2)*exp(a[2,1])*exp(a[6,1]))-6*exp(2*a[6,1]);
    dvdc6[9,1]= -3*a[1,1]*exp(a[2,1])*exp(a[6,1]);

    dvdc7[1,1]= 0;
    dvdc7[2,1]= 0;
    dvdc7[3,1]= 0;
    dvdc7[4,1]= 0;
    dvdc7[5,1]= 0;
    dvdc7[6,1]= 0;
    dvdc7[7,1]= -a[1,1]*exp(a[7,1]);
    dvdc7[8,1]= -(a[1,1]^4)*exp(a[7,1]);
    dvdc7[9,1]= -(a[1,1]^3)*exp(a[7,1]);

    dvdc8[1,1]= -mu2_b(a[5,1],a[8,1]);
    dvdc8[2,1]= 0;
    dvdc8[3,1]= 0;
    dvdc8[4,1]= 0;
    dvdc8[5,1]= 0;
    dvdc8[6,1]= mu3_b(a[5,1],a[8,1]);
    dvdc8[7,1]= 0;
    dvdc8[8,1]= -6*(((a[1,1]^2)*exp(a[2,1]))+exp(a[6,1]))*mu2_b(a[5,1],a[8,1])-
        mu4_b(a[5,1],a[8,1]);
    dvdc8[9,1]= -3*a[1,1]*exp(a[2,1])*mu2_b(a[5,1],a[8,1]);

    retp(dvdc1~dvdc2~dvdc3~dvdc4~dvdc5~dvdc6~dvdc7~dvdc8);
endp;
@----------------------------------------------------------@

proc grad3(a, neq);

    local dvdc1, dvdc2, dvdc3, dvdc4, dvdc5, dvdc6, dvdc7, dvdc8;

    dvdc1 = zeros(neq,1);
    dvdc2 = zeros(neq,1);
    dvdc3 = zeros(neq,1);
    dvdc4 = zeros(neq,1);
    dvdc5 = zeros(neq,1);
    dvdc6 = zeros(neq,1);
    dvdc7 = zeros(neq,1);
    dvdc8 = zeros(neq,1);

    dvdc1[1,1]= -2*a[1,1]*exp(a[2,1]);
    dvdc1[2,1]= -exp(a[2,1]);
    dvdc1[3,1]= 0;
    dvdc1[4,1]= -2*a[1,1]*a[3,1];
    dvdc1[5,1]= -a[3,1];
    dvdc1[6,1]= -3*(a[1,1]^2)*a[3,1];
    dvdc1[7,1]= -exp(a[7,1])-3*exp(a[2,1])*exp(a[4,1]);
    dvdc1[8,1]= -(4*(a[1,1]^3)*exp(a[7,1]))-12*a[1,1]*exp(a[2,1])*(exp(a[6,1])+mu2(a[5,1],a[8,1]));
    dvdc1[9,1]= -3*(a[1,1]^2)*exp(a[7,1])-3*exp(a[2,1])*(exp(a[6,1])+mu2(a[5,1],a[8,1]));
    dvdc1[10,1]= -2*a[1,1]*(exp(a[7,1])+(exp(a[2,1])*exp(a[4,1])));

    dvdc2[1,1]= -(a[1,1]^2)*exp(a[2,1]);
    dvdc2[2,1]= -a[1,1]*exp(a[2,1]);
    dvdc2[3,1]= -exp(a[2,1]);
    dvdc2[4,1]= 0;
    dvdc2[5,1]= 0;
    dvdc2[6,1]= 0;
    dvdc2[7,1]= -3*a[1,1]*exp(a[4,1])*exp(a[2,1]);
    dvdc2[8,1]= -(6*(a[1,1]^2)*(exp(a[6,1]) + mu2(a[5,1],a[8,1])))*exp(a[2,1]);
    dvdc2[9,1]= -(3*a[1,1]*(exp(a[6,1])+mu2(a[5,1],a[8,1])))*exp(a[2,1]);
    dvdc2[10,1]= -((a[1,1]^2)*exp(a[4,1])*exp(a[2,1]))-(exp(a[6,1])+mu2(a[5,1],a[8,1]))*exp(a[2,1]);


    dvdc3[1,1]= 0;
    dvdc3[2,1]= 0;
    dvdc3[3,1]= 0;
    dvdc3[4,1]= -(a[1,1]^2);
    dvdc3[5,1]= -a[1,1];
    dvdc3[6,1]= -(a[1,1]^3);
    dvdc3[7,1]= 0;
    dvdc3[8,1]= 0;
    dvdc3[9,1]= 0;
    dvdc3[10,1]= 0;

    dvdc4[1,1]= 0;
    dvdc4[2,1]= 0;
    dvdc4[3,1]= -exp(a[4,1]);
    dvdc4[4,1]= 0;
    dvdc4[5,1]= 0;
    dvdc4[6,1]= 0;
    dvdc4[7,1]= -3*a[1,1]*exp(a[2,1])*exp(a[4,1]);
    dvdc4[8,1]= 0;
    dvdc4[9,1]= 0;
    dvdc4[10,1]= -((a[1,1]^2)*exp(a[2,1])*exp(a[4,1]))-(exp(a[6,1])+mu2(a[5,1],a[8,1]))*exp(a[4,1]);

    dvdc5[1,1]= -mu2_a(a[5,1],a[8,1]);
    dvdc5[2,1]= 0;
    dvdc5[3,1]= 0;
    dvdc5[4,1]= 0;
    dvdc5[5,1]= 0;
    dvdc5[6,1]= mu3_a(a[5,1],a[8,1]);
    dvdc5[7,1]= 0;
    dvdc5[8,1]= -6*(((a[1,1]^2)*exp(a[2,1]))+exp(a[6,1]))*mu2_a(a[5,1],a[8,1])
        -mu4_a(a[5,1],a[8,1]);
    dvdc5[9,1]= -3*a[1,1]*exp(a[2,1])*mu2_a(a[5,1],a[8,1]);
    dvdc5[10,1]= -(exp(a[2,1])+exp(a[4,1]))*mu2_a(a[5,1],a[8,1]);

    dvdc6[1,1]= -exp(a[6,1]);
    dvdc6[2,1]= 0;
    dvdc6[3,1]= 0;
    dvdc6[4,1]= 0;
    dvdc6[5,1]= 0;
    dvdc6[6,1]= 0;
    dvdc6[7,1]= 0;
    dvdc6[8,1]= -6*mu2(a[5,1],a[8,1])*exp(a[6,1])-(6*(a[1,1]^2)*exp(a[2,1])*exp(a[6,1]))-6*exp(2*a[6,1]);
    dvdc6[9,1]= -3*a[1,1]*exp(a[2,1])*exp(a[6,1]);
    dvdc6[10,1]= -(exp(a[2,1])+exp(a[4,1]))*exp(a[6,1]);

    dvdc7[1,1]= 0;
    dvdc7[2,1]= 0;
    dvdc7[3,1]= 0;
    dvdc7[4,1]= 0;
    dvdc7[5,1]= 0;
    dvdc7[6,1]= 0;
    dvdc7[7,1]= -a[1,1]*exp(a[7,1]);
    dvdc7[8,1]= -(a[1,1]^4)*exp(a[7,1]);
    dvdc7[9,1]= -(a[1,1]^3)*exp(a[7,1]);
    dvdc7[10,1]= -(a[1,1]^2)*exp(a[7,1]);

     dvdc8[1,1]= -mu2_b(a[5,1],a[8,1]);
    dvdc8[2,1]= 0;
    dvdc8[3,1]= 0;
    dvdc8[4,1]= 0;
    dvdc8[5,1]= 0;
    dvdc8[6,1]= mu3_b(a[5,1],a[8,1]);
    dvdc8[7,1]= 0;
    dvdc8[8,1]= -6*(((a[1,1]^2)*exp(a[2,1]))+exp(a[6,1]))*mu2_b(a[5,1],a[8,1])-
        mu4_b(a[5,1],a[8,1]);
    dvdc8[9,1]= -3*a[1,1]*exp(a[2,1])*mu2_b(a[5,1],a[8,1]);
    dvdc8[10,1]= -(exp(a[2,1])+exp(a[4,1]))*mu2_b(a[5,1],a[8,1]);

    retp(dvdc1~dvdc2~dvdc3~dvdc4~dvdc5~dvdc6~dvdc7~dvdc8);
endp;



@  ------------------------------------------------------- @
@ Subroutine to compute squeezes.

  This subroutine compares the values of the objective function at the
  points c+s*dc and c+0.5*s*dc, with dc = the proposed change in the vector
  of parameters, and with step length s initially at 1. s is halved until
  minus the objective function stops declining.
@
proc (2) = squeeze(s_c,s_dc, estim, neq, w, obj);
    local s_c1, s_lm, s_itr, lc1, s_c2, lc2, s_f1, s_f2;


    s_c1=s_c - s_dc; s_lm=1/2; s_itr=1;
    s_f1 = deff(s_c1,Emom, estim, neq);
    lc1 = s_f1*w*s_f1';
    clear s_f1;
    do until s_itr > maxsqez;

        s_c2=s_c-s_lm*s_dc;
        s_f2 = deff(s_c2,Emom, estim, neq);
        lc2 = s_f2*w*s_f2';
        clear s_f2;

        if lc1 <= lc2 and lc1 <= obj;
            retp(s_c1,s_itr-1); goto eoproc;
        else;
            s_c1=s_c2; s_lm=s_lm/2; lc1=lc2;
            s_itr=s_itr+1;
        endif;

    endo;
    retp(s_c2,s_itr-1);

    eoproc:
endp;


@  ------------------------------------------------------- @


@----------Subroutine to trap inversion errors-----------------@

proc invit(w);
local onemore,ww;
onemore = 1;
winv:
trap 1;
  ww = invpd(w);
trap 0;
  if scalerr(ww) > 0;
      if onemore == 1;
        call sysstate(14,1e-128);
        onemore = 0;
        goto winv;
      else;
        ww = -999999;
      endif;
  endif;

retp(ww);

endp;

@----------Subroutine to trap solving errors-----------------@

proc solvit(a,b, estim);
    local dc,onemore;
    onemore = 1;
    solveit:
    trap 1;
/* HJW
    print "estim is " estim;
    print " a is " a;
    print "  b is " b;
*/
    dc = solpd(a,b);
    trap 0;
    if scalerr(dc) > 0;
        @print "a is " a "  b is " b;@
        @print "press any key...."; waitc;@
        if onemore == 1;
            call sysstate(14,1e-128);
            onemore = 0;
            goto solveit;
        else;
            dc = -999999;
        endif;
    endif;

    retp(dc);
endp;



@system;@
